#ifndef KAHAN_H_INCLUDED
#define KAHAN_H_INCLUDED

#include <vector>
#include <numeric>               // accumulate

//  Kahan summation:
// https://en.wikipedia.org/wiki/Kahan_summation_algorithm
// with accumulation: https://stackoverflow.com/questions/10330002/sum-of-small-double-numbers-c
//     https://github.com/JuliaLang/julia/issues/199
// Examples/Reihe_Kahan.cpp

// parallel
//      https://www.physicsforums.com/threads/parallel-kahan-summation.716446/
//      http://onlinelibrary.wiley.com/doi/10.1002/scj.4690240706/abstract
//      https://arxiv.org/pdf/1401.0248.pdf
//      https://hal.archives-ouvertes.fr/hal-00477511/document
//      http://www.drdobbs.com/floating-point-summation/184403224

//     David Goldberg. "What Every Computer Scientist Should Know About Floating-Point Arithmetic,'' ACM Computing Surveys, Vol. 23, #1, March 1991, pp. 5-48.
//     http://www.itu.dk/~sestoft/bachelor/IEEE754_article.pdf


//! @brief Sum of all vector elements.
//!        Summation is performed via the <a href="https://en.wikipedia.org/wiki/Kahan_summation_algorithm">Kahan summation</a>.
//!
//! @param[in]  v vector
//! @return     sum
//!

template<class T>
T KahanSum(const std::vector<T> &v)
{
    T sum = T(0.0),
        c = T(0.0);

    for (size_t i=0; i<v.size(); ++i)
    {
        T y = v[i] -c;
        T t = sum + y;
        c = (t-sum) - y;
        sum = t;
    }

    return sum;
}

//! @brief Sum of all vector elements performed in ascending index order.
//!
//! @param[in]  vector
//! @return     sum
//!
template<class T>
T SumForw(const std::vector<T> &v)
{
    T sum = T(0.0);

    for (std::size_t i=0; i<v.size(); ++i)
    {
        sum += v[i];
    }

    return sum;
}

//! @brief Sum of all vector elements performed in descending index order.
//!
//! @param[in]  vector
//! @return     sum
//!
template<class T>
T SumBackw(const std::vector<T> &v)
{
    T sum = T(0.0);

    //http://en.cppreference.com/w/cpp/types/size_t
    for (std::size_t i=v.size(); i-- >0;)
    {
        sum += v[i];
    }

    return sum;
}

//! @brief Sum of all vector elements via accumulate.
//!        Contains also round off errors.
//!
//! @param[in]  vector
//! @return     sum
//!
template<class T>
T SumAccu(const std::vector<T> &v)
{
    return std::accumulate(v.cbegin(),end(v),T(0));
}


#endif // KAHAN_H_INCLUDED
