//  g++ -O3 -std=c++14 main.cpp  -fext-numeric-literals main.cpp -lgmpxx -lgmp  -lquadmath
//  codeblocks: option -std=c++14 has to be set globally
//
// g++, clang++, icpc, pgc++ equal timing
// '-O3' vs. '-Ofast'  (accuracy/speed)

// using mpf:
//  g++ -O3 -std=c++14 main.cpp  -lgmpxx -lgmp
//
// using float128,   see Ideen/gmp/float128.cpp
//   g++ -O3 -std=c++14  -fext-numeric-literals main.cpp -lquadmath
// or, everything
//   g++ -O3 -std=c++14  -fext-numeric-literals main.cpp -lgmpxx -lgmp  -lquadmath

#include <iostream>
#include <typeinfo>                     // operator typeid
#include <cstdlib>                      // needed for atoi, exit
#include <cmath>
#include <limits>
#include <numeric>                      // iota
#include <algorithm>                    // for_each

//   http://www.boost.org/doc/libs/1_66_0/libs/multiprecision/doc/html/boost_multiprecision/tut/floats/gmp_float.html
#include <boost/multiprecision/gmp.hpp>          // mpf_float_50, mpf_float
#include <boost/multiprecision/float128.hpp>     // float128
using namespace boost::multiprecision;

#include "kahan.h"

using namespace std;

template <class T>
void compare_summation(const std::vector<T> &v)
{
    cout << "######################################################\n";
    //		Print the index where cancellation has to be expected
    cout << "Type is: " << typeid(T).name() << '\n';
    cout << "The first sum will be rather precise until  n = "
         << ceil(sqrt(1./numeric_limits<T>::epsilon() )) << endl << endl;

    double tstart;

    //		Calculate in ascending order
    tstart = clock();
    auto   s1 = SumForw(v);
    double t1 = (clock()-tstart)/CLOCKS_PER_SEC;

//		Calculate in descending order
    tstart = clock();
    auto   s2 = SumBackw(v);
    double t2 = (clock()-tstart)/CLOCKS_PER_SEC;

//  Calculate in ascending order using Kahan summation
    tstart = clock();
    auto   s3 = KahanSum(v);
    double t3 = (clock()-tstart)/CLOCKS_PER_SEC;

//    const auto M6=M_PI*M_PI/6;
    const auto M6=T(M_PIl)*T(M_PIl)/T(6.0);
    cout << "                 asc            desc            Kahan" << endl;
    cout << "value          " << s1 << "        " << s2 << "         " << s3 << endl;
    cout << "Pi*Pi/6-s    " << M6-s1 << "    " << M6-s2 << "     " << M6-s3 << endl;
    cout << "diff                " << s2-s1 << "       " << s3-s2 << endl;
    cout << "time [sec.]    "  << t1 << "       " << t2 << "        " << t3 << endl;

    return;
}

//! @brief Initializes vector element @p v_k with 1/((k+1)*(k+1)).
//!
//! @param[in,out]  v vector
//! @return
//!
//!
template <class T>
void init(std::vector<T> &v)
{
//           http://en.cppreference.com/w/cpp/algorithm/iota
    std::iota(v.begin(),v.end(),T(1.0));     // 1, 2, 3, ...
//           http://en.cppreference.com/w/cpp/algorithm/for_each
//    std::for_each(v.begin(), v.end(), [](auto &p)      // auto: C++-14
    std::for_each(v.begin(), v.end(), [](T &p)      // auto: C++-14
    {
        p=T(1.0)/(p*p);
    });
    return;
}

int main(const int argc, const char *argv[])
{
    int n;
    if (argc > 1)                       // at least one argument
    {
        n = atoi(argv[1]);              // atoi : ACSII to Integer
    }
    else                                // standard input
    {
        n = 6500000;
//        n = 650000000;
//        n = 10;
    }

    vector<float> v1(n);
    init(v1);
    compare_summation(v1);

    vector<mpf_float_50> v2(n);
    init(v2);
    compare_summation(v2);

    vector<float128> v3(n);
    init(v3);
    compare_summation(v3);

    return 0;
}

